#pragma once
#include "Exception.h"
#include <iostream>
#include <cmath>
#include <fstream>
#include <vector>
#include <ranges>
#include <iterator>
#include <algorithm>

//1
void createDataFileStream(const std::string& a_filename)
{
	std::ofstream out(a_filename);
	if (!out.is_open()) 
	{
		throw FileError();
	}
	for (int i: std::ranges::iota_view(0, 100))
	{
		out << rand() % 101 - 50 << "\n";
	}
}

template<class T>
void createDataGenerate(const std::string& a_filename)
{
	std::ofstream out(a_filename);
	if (!out.is_open()) 
	{
		throw FileError();
	}

	using Type = typename  T::value_type;
	T temp(100);

	std::generate(temp.begin(), temp.end(), []() {return rand() % 101 - 50; });
	std::copy(temp.begin(), temp.end(), std::ostream_iterator<Type>(out, "\n"));
}

//2
template<class T>
T loadDataFileStream(const std::string& a_fileName)
{
	T res;
	std::ifstream file((a_fileName));
	if (!file.is_open()) 
	{
		throw FileError();
	}

	using Type = typename  T::value_type;
	Type buff;

	while (!file.eof())
	{
		file >> buff;
		res.push_back(buff);
	}
	return res;
}


template<class T>
T loadDataCopy(const std::string& a_fileName)
{
	using Type = typename  T::value_type;
	T res(100);

	std::ifstream file(a_fileName);
	if (!file.is_open()) 
	{
		throw FileError();
	}

	std::istream_iterator<Type> begin(file);
	std::istream_iterator<Type> end;
	
	std::copy(begin, end, res.begin());
	return res;
}

//3
//     
template<class T>
void Modify(T& a_container)
{
	const auto maxElem = std::max_element(a_container.begin(), a_container.end());

	std::for_each(a_container.begin(), a_container.end(),
		[maxElem](auto& element) {element -= *maxElem; });

	using Type = typename  T::value_type;
	Type sum = 0;
	std::for_each(a_container.begin(), a_container.end(), 
		[&sum](const auto& element) 
		{
			sum += element; 
		});
	a_container.push_back(sum);
	a_container.push_back(abs(sum / 100));
}

template<class T, class F>
void ModifyForEach(T a_container, F a_func)
{
	std::for_each(a_container.begin(), a_container.end(), a_func);
}

template<class Iter>
void ModifyByIters(Iter a_begin, Iter a_end)
{
	const auto maxElem = std::max_element(a_begin, a_end);

	while (a_begin != a_end)
	{
		*a_begin -= *maxElem;
		++a_begin;
	}
}

template<class T>
void ModifyTransform(T& a_container)
{
	const auto maxElem = std::max_element(a_container.begin(), a_container.end());
	std::transform(a_container.begin(), a_container.end(), a_container.begin(),
		[maxElem](auto& elem) 
		{
			return elem -= *maxElem;
		});
}

//4
template<class T>
void outputResultFileSteam(const T& a_container, const std::string& a_fileName)
{
	std::ofstream file(a_fileName, std::ios::trunc);

	if (!file.is_open())
	{
		throw FileError();
	}

	for (const auto& i : a_container)
	{
		file << i << std::endl;
	}
}

template<class T>
void outputResultCopy(const T& a_container, const std::string& a_fileName)
{
	std::ofstream file(a_fileName, std::ios::trunc);

	if (!file.is_open()) 
	{
		throw FileError();
	}

	std::copy(a_container.begin(), a_container.end(), std::ostream_iterator<float>(file, "\n"));
}


void testPart1()
{
	const std::string file1 = "file1.txt";
	const std::string file2 = "file2.txt";
	std::vector<float> container;

	bool flag = true;
	while (flag)
	{	
		int select;
		std::cout << "1. Create Data by File Streams \n2. Create Data by iterators\n" << 
			         "3. Load Data by File Stream    \n4. Load Data by Copy()\n" <<
			         "5. Modify                      \n6. Modify by for_each()\n" <<
			         "7. Modify with Iterators       \n8. Modify with Transform()\n" <<
			         "9: Output Result by File Stream\n10.Output Result by iterators\n" << 
			         "11.Exit\n";

		
		std::cout << "Select option:" << std::endl;
		std::cin >> select;
		
		try
		{
			switch (select)
			{
			case 1: 
				createDataFileStream("file1.txt"); 
				break;
			case 2: 
				createDataGenerate<std::vector<float>>("file1.txt");
				break;
			case 3: 
				container = loadDataFileStream<std::vector<float>>("file1.txt") ;
				break;
			case 4: 
				container = loadDataCopy<std::vector<float>>("file1.txt");
				break;
			case 5: 
				Modify<std::vector<float>>(container);
				break;
			case 6: 
				ModifyForEach<std::vector<float>>(container, 
					[container](float& element) 
					{
						return element -= *std::max_element(container.begin(), container.end());
					});
				break;
			case 7: 
				ModifyByIters(container.begin(), container.end());
				break;
			case 8: 
				ModifyTransform(container);
				break;
			case 9: 
				outputResultFileSteam(container, "file2.txt");
				break;
			case 10: 
				outputResultCopy(container, "file2.txt");
				break;
			case 11:
				flag = false; 
				break;
			default: 
				std::cout << "ERROR! Try again" << std::endl;
			}
		}
		catch (std::exception& error)
		{
			std::cerr << error.what() << std::endl;
		}
	}


}